@@ -266,7 +266,7 @@ gjs_array_to_g_list(JSContext   *context,
     }
 
     for (i = 0; i < length; ++i) {
-        GArgument elem_arg;
+        GArgument elem_arg = { 0 };
 
         elem = JSVAL_VOID;
         if (!JS_GetElement(context, JSVAL_TO_OBJECT(array_value),
@@ -356,7 +356,7 @@ gjs_object_to_g_hash(JSContext   *context,
 
    while (!JSID_IS_VOID(prop_id)) {
         jsval key_js, val_js;
-        GArgument key_arg, val_arg;
+        GArgument key_arg = { 0 }, val_arg = { 0 };
 
         if (!JS_IdToValue(context, prop_id, &key_js))
             goto free_hash_and_fail;
@@ -481,22 +481,21 @@ gjs_array_to_strv(JSContext   *context,
 static JSBool
 gjs_string_to_intarray(JSContext   *context,
                        jsval        string_val,
-                       GITypeInfo  *param_info,
-                       void       **arr_p)
+                       GITypeTag    param_tag,
+                       void       **arr_p,
+                       gsize       *length)
 {
     GITypeTag element_type;
     char *result;
     guint16 *result16;
-    gsize length;
 
-    element_type = g_type_info_get_tag(param_info);
-    element_type = replace_gtype(element_type);
+    element_type = replace_gtype(param_tag);
 
     switch (element_type) {
     case GI_TYPE_TAG_INT8:
     case GI_TYPE_TAG_UINT8:
         if (!gjs_string_get_binary_data(context, string_val,
-                                        &result, &length))
+                                        &result, length))
             return JS_FALSE;
         *arr_p = result;
         return JS_TRUE;
@@ -504,15 +503,15 @@ gjs_string_to_intarray(JSContext   *context,
     case GI_TYPE_TAG_INT16:
     case GI_TYPE_TAG_UINT16:
         if (!gjs_string_get_uint16_data(context, string_val,
-                                        &result16, &length))
+                                        &result16, length))
             return JS_FALSE;
         *arr_p = result16;
         return JS_TRUE;
 
     default:
-        // can't convert a string to this type.
-        gjs_throw(context, "Cannot convert string to type '%s'",
-                  g_base_info_get_name((GIBaseInfo*) param_info));
+        /* can't convert a string to this type */
+        gjs_throw(context, "Cannot convert string to array of '%s'",
+                  g_type_tag_to_string (element_type));
         return JS_FALSE;
     }
 }
@@ -527,10 +526,11 @@ gjs_array_to_intarray(JSContext   *context,
 {
     /* nasty union types in an attempt to unify the various int types */
     union { guint32 u; gint32 i; } intval;
-    union { guint8 u8[0]; guint16 u16[0]; guint32 u32[0]; } *result;
+    void *result;
     unsigned i;
 
-    result = g_malloc0(length * intsize);
+    /* add one so we're always zero terminated */
+    result = g_malloc0((length+1) * intsize);
 
     for (i = 0; i < length; ++i) {
         jsval elem;
@@ -560,11 +560,11 @@ gjs_array_to_intarray(JSContext   *context,
         /* Note that this is truncating assignment. */
         switch (intsize) {
         case 1:
-            result->u8[i] = (gint8) intval.u; break;
+            ((guint8*)result)[i] = (gint8) intval.u; break;
         case 2:
-            result->u16[i] = (gint16) intval.u; break;
+            ((guint16*)result)[i] = (gint16) intval.u; break;
         case 4:
-            result->u32[i] = (gint32) intval.u; break;
+            ((guint32*)result)[i] = (gint32) intval.u; break;
         default:
             g_assert_not_reached();
         }
@@ -575,18 +575,134 @@ gjs_array_to_intarray(JSContext   *context,
     return JS_TRUE;
 }
 
+static JSBool
+gjs_array_to_floatarray(JSContext   *context,
+                        jsval        array_value,
+                        unsigned int length,
+                        void       **arr_p,
+                        gboolean     is_double)
+{
+    unsigned int i;
+    void *result;
+
+    /* add one so we're always zero terminated */
+    result = g_malloc0((length+1) * (is_double ? sizeof(double) : sizeof(float)));
+
+    for (i = 0; i < length; ++i) {
+        jsval elem;
+        jsdouble val;
+        JSBool success;
+
+        elem = JSVAL_VOID;
+        if (!JS_GetElement(context, JSVAL_TO_OBJECT(array_value),
+                           i, &elem)) {
+            g_free(result);
+            gjs_throw(context,
+                      "Missing array element %u",
+                      i);
+            return JS_FALSE;
+        }
+
+        /* do whatever sign extension is appropriate */
+        success = JS_ValueToNumber(context, elem, &val);
+
+        if (!success) {
+            g_free(result);
+            gjs_throw(context,
+                      "Invalid element in array");
+            return JS_FALSE;
+        }
+
+        /* Note that this is truncating assignment. */
+        if (is_double) {
+            double *darray = (double*)result;
+            darray[i] = val;
+        } else {
+            float *farray = (float*)result;
+            farray[i] = val;
+        }
+    }
+
+    *arr_p = result;
+
+    return JS_TRUE;
+}
+
+static JSBool
+gjs_array_to_ptrarray(JSContext   *context,
+                      jsval        array_value,
+                      unsigned int length,
+                      GITransfer   transfer,
+                      GITypeInfo  *param_info,
+                      void       **arr_p)
+{
+    unsigned int i;
+
+    /* Always one extra element, to cater for null terminated arrays */
+    void **array = g_malloc((length + 1) * sizeof(gpointer));
+    array[length] = NULL;
+
+    for (i = 0; i < length; i++) {
+        jsval elem;
+        GIArgument arg;
+        arg.v_pointer = NULL;
+
+        JSBool success;
+
+        elem = JSVAL_VOID;
+        if (!JS_GetElement(context, JSVAL_TO_OBJECT(array_value),
+                           i, &elem)) {
+            g_free(array);
+            gjs_throw(context,
+                      "Missing array element %u",
+                      i);
+            return JS_FALSE;
+        }
+
+        success = gjs_value_to_g_argument (context,
+                                           elem,
+                                           param_info,
+                                           NULL, /* arg name */
+                                           GJS_ARGUMENT_ARRAY_ELEMENT,
+                                           transfer,
+                                           FALSE, /* absent better information, FALSE for now */
+                                           &arg);
+
+        if (!success) {
+            g_free(array);
+            gjs_throw(context,
+                      "Invalid element in array");
+            return JS_FALSE;
+        }
+
+        array[i] = arg.v_pointer;
+    }
+
+    *arr_p = array;
+    return JS_TRUE;
+} 
+
 static JSBool
 gjs_array_to_array(JSContext   *context,
                    jsval        array_value,
-                   unsigned int length,
+                   gsize        length,
+                   GITransfer   transfer,
+                   GITypeTag    param_tag,
                    GITypeInfo  *param_info,
                    void       **arr_p)
 {
     enum { UNSIGNED=FALSE, SIGNED=TRUE };
     GITypeTag element_type;
 
-    element_type = g_type_info_get_tag(param_info);
-    element_type = replace_gtype(element_type);
+    element_type = replace_gtype(param_tag);
+
+    if (element_type == GI_TYPE_TAG_INTERFACE) {
+        GIBaseInfo *interface_info = g_type_info_get_interface(param_info);
+        GIInfoType info_type = g_base_info_get_type(interface_info);
+        if (info_type == GI_INFO_TYPE_ENUM || info_type == GI_INFO_TYPE_FLAGS)
+            element_type = g_enum_info_get_storage_type ((GIEnumInfo*) interface_info);
+        g_base_info_unref(interface_info);
+    }
 
     switch (element_type) {
     case GI_TYPE_TAG_UTF8:
@@ -609,6 +725,25 @@ gjs_array_to_array(JSContext   *context,
     case GI_TYPE_TAG_INT32:
         return gjs_array_to_intarray
             (context, array_value, length, arr_p, 4, SIGNED);
+    case GI_TYPE_TAG_FLOAT:
+        return gjs_array_to_floatarray
+            (context, array_value, length, arr_p, FALSE);
+    case GI_TYPE_TAG_DOUBLE:
+        return gjs_array_to_floatarray
+            (context, array_value, length, arr_p, TRUE);
+
+    /* Everything else is a pointer type */
+    case GI_TYPE_TAG_INTERFACE:
+    case GI_TYPE_TAG_ARRAY:
+    case GI_TYPE_TAG_GLIST:
+    case GI_TYPE_TAG_GSLIST:
+    case GI_TYPE_TAG_GHASH:
+        return gjs_array_to_ptrarray(context,
+                                     array_value,
+                                     length,
+                                     transfer == GI_TRANSFER_CONTAINER ? GI_TRANSFER_NOTHING : transfer,
+                                     param_info,
+                                     arr_p);
     default:
         gjs_throw(context,
                   "Unhandled array element type %d", element_type);
@@ -616,21 +751,28 @@ gjs_array_to_array(JSContext   *context,
     }
 }
 
-static JSBool
-gjs_array_to_g_array(JSContext   *context,
-                     jsval        array_value,
-                     unsigned int length,
-                     GITypeInfo  *param_info,
-                     void       **arr_p)
+static GArray*
+gjs_g_array_new_for_type(JSContext    *context,
+                         unsigned int  length,
+                         GITypeInfo   *param_info)
 {
-    GArray *array;
     GITypeTag element_type;
-    gpointer contents;
     guint element_size;
 
     element_type = g_type_info_get_tag(param_info);
     element_type = replace_gtype(element_type);
 
+    if (element_type == GI_TYPE_TAG_INTERFACE) {
+        GIInterfaceInfo *interface_info = g_type_info_get_interface(param_info);
+        GIInfoType interface_type = g_base_info_get_type(interface_info);
+
+        if (interface_type == GI_INFO_TYPE_ENUM
+            || interface_type == GI_INFO_TYPE_FLAGS)
+            element_type = g_enum_info_get_storage_type((GIEnumInfo*) interface_info);
+
+        g_base_info_unref((GIBaseInfo*) interface_info);
+    }
+
     switch (element_type) {
     case GI_TYPE_TAG_UINT8:
     case GI_TYPE_TAG_INT8:
@@ -648,44 +790,29 @@ gjs_array_to_g_array(JSContext   *context,
     case GI_TYPE_TAG_INT64:
       element_size = sizeof(guint64);
       break;
+    case GI_TYPE_TAG_FLOAT:
+      element_size = sizeof(gfloat);
+      break;
+    case GI_TYPE_TAG_DOUBLE:
+      element_size = sizeof(gdouble);
+      break;
+    case GI_TYPE_TAG_INTERFACE:
+    case GI_TYPE_TAG_UTF8:
+    case GI_TYPE_TAG_FILENAME:
+    case GI_TYPE_TAG_ARRAY:
+    case GI_TYPE_TAG_GLIST:
+    case GI_TYPE_TAG_GSLIST:
+    case GI_TYPE_TAG_GHASH:
+    case GI_TYPE_TAG_ERROR:
+      element_size = sizeof(gpointer);
+      break;
     default:
         gjs_throw(context,
                   "Unhandled GArray element-type %d", element_type);
-        return JS_FALSE;
+        return NULL;
     }
 
-    /* create a C array */
-    if (!gjs_array_to_array (context,
-                             array_value,
-                             length,
-                             param_info,
-                             &contents))
-      return JS_FALSE;
-
-    /* append that array to the GArray */
-    array = g_array_sized_new(TRUE, TRUE, element_size, length);
-    g_array_append_vals(array, contents, length);
-
-    g_free(contents);
-
-    *arr_p = array;
-
-    return JS_TRUE;
-}
-
-static JSBool
-gjs_array_to_byte_array(JSContext  *context,
-                        jsval       value,
-                        void      **arr_p)
-{
-   GByteArray *byte_array;
-   byte_array = gjs_byte_array_get_byte_array(context,
-                                              JSVAL_TO_OBJECT(value));
-   if (!byte_array)
-       return JS_FALSE;
-
-   *arr_p = byte_array;
-   return JS_TRUE;
+    return g_array_sized_new(TRUE, FALSE, element_size, length);
 }
 
 static gchar *
@@ -703,11 +830,90 @@ get_argument_display_name(const char     *arg_name,
         return g_strdup("List element");
     case GJS_ARGUMENT_HASH_ELEMENT:
         return g_strdup("Hash element");
+    case GJS_ARGUMENT_ARRAY_ELEMENT:
+        return g_strdup("Array element");
     }
 
     g_assert_not_reached ();
 }
 
+static JSBool
+gjs_array_to_explicit_array_internal(JSContext       *context,
+                                     jsval            value,
+                                     GITypeInfo      *type_info,
+                                     const char      *arg_name,
+                                     GjsArgumentType  arg_type,
+                                     GITransfer       transfer,
+                                     gboolean         may_be_null,
+                                     gpointer        *contents,
+                                     gsize           *length_p)
+{
+    JSBool ret = JS_TRUE;
+    GIArrayType array_type = g_type_info_get_array_type(type_info);
+    GITypeInfo *param_info;
+    GITypeTag param_tag;
+
+    param_info = g_type_info_get_param_type(type_info, 0);
+    param_tag = g_type_info_get_tag(param_info);
+
+    if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
+        /* HACK: ByteArrays are reported as having a param of gpointer (GI_TYPE_TAG_VOID),
+           which is obviously wrong. Override it here */
+        param_tag = GI_TYPE_TAG_UINT8;
+    }
+
+    if ((JSVAL_IS_NULL(value) && !may_be_null) ||
+        (!JSVAL_IS_STRING(value) && !JSVAL_IS_OBJECT(value) && !JSVAL_IS_NULL(value))) {
+        gchar *display_name = get_argument_display_name(arg_name, arg_type);
+        gjs_throw(context, "Expected type %s for %s but got type '%s' %p",
+                  g_type_tag_to_string(param_tag),
+                  display_name,
+                  JS_GetTypeName(context,
+                                 JS_TypeOfValue(context, value)),
+                  JSVAL_IS_OBJECT(value) ? JSVAL_TO_OBJECT(value) : NULL);
+        g_free(display_name);
+        return JS_FALSE;
+    }
+
+    if (JSVAL_IS_NULL(value)) {
+        *contents = NULL;
+        *length_p = 0;
+    } else if (JSVAL_IS_STRING(value)) {
+        /* Allow strings as int8/uint8/int16/uint16 arrays */
+        if (!gjs_string_to_intarray(context, value, param_tag,
+                                    contents, length_p))
+            ret = JS_FALSE;
+    } else if (gjs_object_has_property(context,
+                                       JSVAL_TO_OBJECT(value),
+                                       "length")) {
+        jsval length_value;
+        guint32 length;
+
+        if (!gjs_object_require_property(context,
+                                         JSVAL_TO_OBJECT(value), NULL,
+                                         "length",
+                                         &length_value) ||
+            !JS_ValueToECMAUint32(context, length_value, &length)) {
+            ret = JS_FALSE;
+        } else {
+            if (!gjs_array_to_array(context,
+                                    value,
+                                    length,
+                                    transfer,
+                                    param_tag,
+                                    param_info,
+                                    contents))
+                ret = JS_FALSE;
+
+            *length_p = length;
+        }
+    }
+
+    g_base_info_unref((GIBaseInfo*) param_info);
+
+    return ret;
+}
+
 JSBool
 gjs_value_to_g_argument(JSContext      *context,
                         jsval           value,
@@ -949,6 +1155,8 @@ gjs_value_to_g_argument(JSContext      *context,
                     if (transfer != GI_TRANSFER_NOTHING) {
                         if (g_type_is_a(gtype, G_TYPE_BOXED))
                             arg->v_pointer = g_boxed_copy (gtype, arg->v_pointer);
+                        else if (g_type_is_a(gtype, G_TYPE_VARIANT))
+                            g_variant_ref (arg->v_pointer);
                         else {
                             gjs_throw(context,
                                       "Can't transfer ownership of a structure type not registered as boxed");
@@ -1158,87 +1366,64 @@ gjs_value_to_g_argument(JSContext      *context,
         }
         break;
 
-    case GI_TYPE_TAG_ARRAY:
-        if (JSVAL_IS_NULL(value)) {
-            arg->v_pointer = NULL;
-            if (!may_be_null) {
-                wrong = TRUE;
-                report_type_mismatch = TRUE;
+    case GI_TYPE_TAG_ARRAY: {
+        gpointer data;
+        gsize length;
+        GIArrayType array_type = g_type_info_get_array_type(type_info);
+
+        if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY && JSVAL_IS_OBJECT(value)) {
+            GByteArray *byte_array;
+
+            byte_array = gjs_byte_array_get_byte_array(context, JSVAL_TO_OBJECT(value));
+            if (byte_array) {
+                /* extra reference will be removed gjs_g_argument_release
+                   (and actually has no effect given that finalizer invokes
+                   g_byte_array_free without caring of refcount) */
+                arg->v_pointer = g_byte_array_ref(byte_array);
+                break;
             }
-        } else if (JSVAL_IS_STRING(value)) {
-            /* Allow strings as int8/uint8/int16/uint16 arrays */
-            GITypeInfo *param_info;
+            /* otherwise this is not a JS ByteArray, so fall through to extracting
+               all elements from the String or the Array */
+        }
 
-            param_info = g_type_info_get_param_type(type_info, 0);
-            g_assert(param_info != NULL);
-
-            if (!gjs_string_to_intarray(context, value, param_info,
-                                        &arg->v_pointer))
-                wrong = TRUE;
-
-            g_base_info_unref((GIBaseInfo*) param_info);
-        } else if (!JSVAL_IS_OBJECT(value)) {
+        if (!gjs_array_to_explicit_array_internal(context,
+                                                  value,
+                                                  type_info,
+                                                  arg_name,
+                                                  arg_type,
+                                                  transfer,
+                                                  may_be_null,
+                                                  &data,
+                                                  &length)) {
             wrong = TRUE;
-            report_type_mismatch = TRUE;
-        } else if (gjs_object_has_property(context,
-                                           JSVAL_TO_OBJECT(value),
-                                           "length")) {
-            jsval length_value;
-            guint32 length;
+            break;
+        }
 
-            if (!gjs_object_require_property(context,
-                                             JSVAL_TO_OBJECT(value), NULL,
-                                             "length",
-                                             &length_value) ||
-                !JS_ValueToECMAUint32(context, length_value, &length)) {
+        if (array_type == GI_ARRAY_TYPE_C) {
+            arg->v_pointer = data;
+        } else if (array_type == GI_ARRAY_TYPE_ARRAY) {
+            GITypeInfo *param_info = g_type_info_get_param_type(type_info, 0);
+            GArray *array = gjs_g_array_new_for_type(context, length, param_info);
+
+            if (!array)
                 wrong = TRUE;
-            } else {
-                GIArrayType array_type = g_type_info_get_array_type(type_info);
-                GITypeInfo *param_info;
-
-                param_info = g_type_info_get_param_type(type_info, 0);
-                g_assert(param_info != NULL);
-
-                if (array_type == GI_ARRAY_TYPE_C) {
-                    if (!gjs_array_to_array (context,
-                                             value,
-                                             length,
-                                             param_info,
-                                             &arg->v_pointer))
-                      wrong = TRUE;
-                } else if (array_type == GI_ARRAY_TYPE_ARRAY) {
-                    if (!gjs_array_to_g_array (context,
-                                               value,
-                                               length,
-                                               param_info,
-                                               &arg->v_pointer))
-                      wrong = TRUE;
-                } else if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
-                    if (!gjs_array_to_byte_array(context,
-                                                 value,
-                                                 &arg->v_pointer))
-                        wrong = TRUE;
-                /* FIXME: support PtrArray */
-                } else {
-                    g_assert_not_reached();
-                }
-
-                g_base_info_unref((GIBaseInfo*) param_info);
-            }
-        } else {
-            GIArrayType array_type = g_type_info_get_array_type(type_info);
-            if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
-                if (!gjs_array_to_byte_array(context,
-                                             value,
-                                             &arg->v_pointer))
-                    wrong = TRUE;
-            } else {
-                wrong = TRUE;
-                report_type_mismatch = TRUE;
+            else {
+                g_array_append_vals(array, data, length);
+                arg->v_pointer = array;
             }
+
+            g_free(data);
+            g_base_info_unref((GIBaseInfo*) param_info);
+        } else if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
+            GByteArray *byte_array = g_byte_array_sized_new(length);
+
+            g_byte_array_append(byte_array, data, length);
+            arg->v_pointer = byte_array;
+
+            g_free(data);
         }
         break;
-
+    }
     default:
         gjs_debug(GJS_DEBUG_ERROR,
                   "Unhandled type %s for JavaScript to GArgument conversion",
@@ -1408,21 +1593,39 @@ gjs_value_to_arg(JSContext  *context,
                  GArgument  *arg)
 {
     GITypeInfo type_info;
-    gboolean result;
 
     g_arg_info_load_type(arg_info, &type_info);
 
-    result =
-        gjs_value_to_g_argument(context, value,
-                                &type_info,
-                                g_base_info_get_name( (GIBaseInfo*) arg_info),
-                                (g_arg_info_is_return_value(arg_info) ?
-                                 GJS_ARGUMENT_RETURN_VALUE : GJS_ARGUMENT_ARGUMENT),
-                                g_arg_info_get_ownership_transfer(arg_info),
-                                g_arg_info_may_be_null(arg_info),
-                                arg);
+    return gjs_value_to_g_argument(context, value,
+                                   &type_info,
+                                   g_base_info_get_name( (GIBaseInfo*) arg_info),
+                                   (g_arg_info_is_return_value(arg_info) ?
+                                    GJS_ARGUMENT_RETURN_VALUE : GJS_ARGUMENT_ARGUMENT),
+                                   g_arg_info_get_ownership_transfer(arg_info),
+                                   g_arg_info_may_be_null(arg_info),
+                                   arg);
+}
 
-    return result;
+JSBool
+gjs_value_to_explicit_array (JSContext  *context,
+                             jsval       value,
+                             GIArgInfo  *arg_info,
+                             GArgument  *arg,
+                             gsize      *length_p)
+{
+    GITypeInfo type_info;
+
+    g_arg_info_load_type(arg_info, &type_info);
+
+    return gjs_array_to_explicit_array_internal(context,
+                                                value,
+                                                &type_info,
+                                                g_base_info_get_name((GIBaseInfo*) arg_info),
+                                                GJS_ARGUMENT_ARGUMENT,
+                                                g_arg_info_get_ownership_transfer(arg_info),
+                                                g_arg_info_may_be_null(arg_info),
+                                                &arg->v_pointer,
+                                                length_p);
 }
 
 static JSBool
@@ -1491,6 +1694,156 @@ gjs_array_from_g_list (JSContext  *context,
     return result;
 }
 
+static JSBool
+gjs_array_from_carray_internal (JSContext  *context,
+                                jsval      *value_p,
+                                GITypeInfo *param_info,
+                                guint       length,
+                                gpointer    array)
+{
+    JSObject *obj;
+    jsval elem;
+    GArgument arg;
+    JSBool result;
+    GITypeTag element_type;
+    guint i;
+
+    element_type = g_type_info_get_tag(param_info);
+    element_type = replace_gtype(element_type);
+
+    /* Special case array(guint8) */
+    if (element_type == GI_TYPE_TAG_UINT8) {
+        GByteArray gbytearray;
+
+        gbytearray.data = array;
+        gbytearray.len = length;
+        
+        obj = gjs_byte_array_from_byte_array (context, &gbytearray);
+        if (obj == NULL)
+            return JS_FALSE;
+        *value_p = OBJECT_TO_JSVAL(obj);
+        return JS_TRUE;
+    } 
+
+    obj = JS_NewArrayObject(context, 0, NULL);
+    if (obj == NULL)
+      return JS_FALSE;
+
+    *value_p = OBJECT_TO_JSVAL(obj);
+
+    elem = JSVAL_VOID;
+    JS_AddValueRoot(context, &elem);
+
+#define ITERATE(type) \
+    for (i = 0; i < length; i++) { \
+        arg.v_##type = *(((g##type*)array) + i);                         \
+        if (!gjs_value_from_g_argument(context, &elem, param_info, &arg)) \
+          goto finally; \
+        if (!JS_DefineElement(context, obj, i, elem, NULL, NULL, \
+              JSPROP_ENUMERATE)) \
+          goto finally; \
+    }
+
+    switch (element_type) {
+        case GI_TYPE_TAG_UINT8:
+          ITERATE(uint8);
+          break;
+        case GI_TYPE_TAG_INT8:
+          ITERATE(int8);
+          break;
+        case GI_TYPE_TAG_UINT16:
+          ITERATE(uint16);
+          break;
+        case GI_TYPE_TAG_INT16:
+          ITERATE(int16);
+          break;
+        case GI_TYPE_TAG_UINT32:
+          ITERATE(uint32);
+          break;
+        case GI_TYPE_TAG_INT32:
+          ITERATE(int32);
+          break;
+        case GI_TYPE_TAG_UINT64:
+          ITERATE(uint64);
+          break;
+        case GI_TYPE_TAG_INT64:
+          ITERATE(int64);
+          break;
+        case GI_TYPE_TAG_FLOAT:
+          ITERATE(float);
+          break;
+        case GI_TYPE_TAG_DOUBLE:
+          ITERATE(double);
+          break;
+        case GI_TYPE_TAG_UTF8:
+        case GI_TYPE_TAG_FILENAME:
+        case GI_TYPE_TAG_ARRAY:
+        case GI_TYPE_TAG_INTERFACE:
+        case GI_TYPE_TAG_GLIST:
+        case GI_TYPE_TAG_GSLIST:
+        case GI_TYPE_TAG_GHASH:
+          ITERATE(pointer);
+          break;
+        default:
+          gjs_throw(context, "Unknown Array element-type %d", element_type);
+          goto finally;
+    }
+
+#undef ITERATE
+
+    result = JS_TRUE;
+
+finally:
+    JS_RemoveValueRoot(context, &elem);
+
+    return result;
+}
+
+static JSBool
+gjs_array_from_carray (JSContext  *context,
+                       jsval      *value_p,
+                       GITypeInfo *type_info,
+                       gpointer    array)
+{
+    gint length;
+    GITypeInfo *param_info;
+    JSBool res;
+
+    length = g_type_info_get_array_fixed_size(type_info);
+
+    if (length == -1) {
+        gjs_throw(context, "FIXME: Only supporting fixed size ARRAYs");
+        return JS_FALSE;
+    }
+
+    param_info = g_type_info_get_param_type(type_info, 0);
+
+    res = gjs_array_from_carray_internal(context, value_p, param_info, length, array);
+
+    g_base_info_unref((GIBaseInfo*)param_info);
+
+    return res;
+}
+
+JSBool
+gjs_value_from_explicit_array(JSContext  *context,
+                              jsval      *value_p,
+                              GITypeInfo *type_info,
+                              GArgument  *arg,
+                              int         length)
+{
+    GITypeInfo *param_info;
+    JSBool res;
+
+    param_info = g_type_info_get_param_type(type_info, 0);
+
+    res = gjs_array_from_carray_internal(context, value_p, param_info, length, arg->v_pointer);
+
+    g_base_info_unref((GIBaseInfo*)param_info);
+
+    return res;
+}
+
 static JSBool
 gjs_array_from_g_array (JSContext  *context,
                         jsval      *value_p,
@@ -1580,6 +1933,114 @@ finally:
     return result;
 }
 
+static JSBool
+gjs_array_from_zero_terminated_c_array (JSContext  *context,
+                                        jsval      *value_p,
+                                        GITypeInfo *param_info,
+                                        gpointer    c_array)
+{
+    JSObject *obj;
+    jsval elem;
+    GArgument arg;
+    JSBool result;
+    GITypeTag element_type;
+    guint i;
+
+    element_type = g_type_info_get_tag(param_info);
+    element_type = replace_gtype(element_type);
+
+    /* Special case array(guint8) */
+    if (element_type == GI_TYPE_TAG_UINT8) {
+        GByteArray gbytearray;
+
+        gbytearray.data = c_array;
+        gbytearray.len = strlen(c_array);
+
+        obj = gjs_byte_array_from_byte_array (context, &gbytearray);
+        if (obj == NULL)
+            return JS_FALSE;
+        *value_p = OBJECT_TO_JSVAL(obj);
+        return JS_TRUE;
+    } 
+
+    obj = JS_NewArrayObject(context, 0, NULL);
+    if (obj == NULL)
+      return JS_FALSE;
+
+    *value_p = OBJECT_TO_JSVAL(obj);
+
+    elem = JSVAL_VOID;
+    JS_AddValueRoot(context, &elem);
+
+#define ITERATE(type) \
+    do { \
+        g##type *array = c_array; \
+        for (i = 0; array[i]; i++) { \
+            arg.v_##type = array[i]; \
+            if (!gjs_value_from_g_argument(context, &elem, param_info, &arg)) \
+                goto finally; \
+            if (!JS_DefineElement(context, obj, i, elem, NULL, NULL, \
+                                  JSPROP_ENUMERATE)) \
+                goto finally; \
+        } \
+    } while(0);
+
+    switch (element_type) {
+        case GI_TYPE_TAG_UINT8:
+          ITERATE(uint8);
+          break;
+        case GI_TYPE_TAG_INT8:
+          ITERATE(int8);
+          break;
+        case GI_TYPE_TAG_UINT16:
+          ITERATE(uint16);
+          break;
+        case GI_TYPE_TAG_INT16:
+          ITERATE(int16);
+          break;
+        case GI_TYPE_TAG_UINT32:
+          ITERATE(uint32);
+          break;
+        case GI_TYPE_TAG_INT32:
+          ITERATE(int32);
+          break;
+        case GI_TYPE_TAG_UINT64:
+          ITERATE(uint64);
+          break;
+        case GI_TYPE_TAG_INT64:
+          ITERATE(int64);
+          break;
+        case GI_TYPE_TAG_FLOAT:
+          ITERATE(float);
+          break;
+        case GI_TYPE_TAG_DOUBLE:
+          ITERATE(double);
+          break;
+        case GI_TYPE_TAG_UTF8:
+        case GI_TYPE_TAG_FILENAME:
+        case GI_TYPE_TAG_ARRAY:
+        case GI_TYPE_TAG_INTERFACE:
+        case GI_TYPE_TAG_GLIST:
+        case GI_TYPE_TAG_GSLIST:
+        case GI_TYPE_TAG_GHASH:
+          ITERATE(pointer);
+          break;
+        default:
+          gjs_throw(context, "Unknown element-type %d", element_type);
+          goto finally;
+    }
+
+#undef ITERATE
+
+    result = JS_TRUE;
+
+finally:
+    JS_RemoveValueRoot(context, &elem);
+
+    return result;
+}
+
+
 static JSBool
 gjs_object_from_g_hash (JSContext  *context,
                         jsval      *value_p,
@@ -1873,29 +2334,21 @@ gjs_value_from_g_argument (JSContext  *context,
 
             if (g_type_info_is_zero_terminated(type_info)) {
                 GITypeInfo *param_info;
-                GITypeTag param_tag;
                 JSBool result;
 
                 param_info = g_type_info_get_param_type(type_info, 0);
                 g_assert(param_info != NULL);
 
-                param_tag = g_type_info_get_tag((GITypeInfo*) param_info);
-
-                if (param_tag == GI_TYPE_TAG_UTF8) {
-                    result = gjs_array_from_strv(context,
-                                                 value_p,
-                                                 arg->v_pointer);
-                } else {
-                    gjs_throw(context, "FIXME: Only supporting null-terminated arrays of strings");
-                    result = FALSE;
-                }
+                result = gjs_array_from_zero_terminated_c_array(context,
+                                                                value_p,
+                                                                param_info,
+                                                                arg->v_pointer);
 
                 g_base_info_unref((GIBaseInfo*) param_info);
 
                 return result;
             } else {
-                gjs_throw(context, "FIXME: Only supporting zero-terminated ARRAYs");
-                return JS_FALSE;
+                return gjs_array_from_carray(context, value_p, type_info, arg->v_pointer);
             }
         } else if (g_type_info_get_array_type(type_info) == GI_ARRAY_TYPE_BYTE_ARRAY) {
             JSObject *array = gjs_byte_array_from_byte_array(context,
@@ -2047,6 +2500,7 @@ gjs_g_arg_release_internal(JSContext  *context,
     case GI_TYPE_TAG_UINT64:
     case GI_TYPE_TAG_FLOAT:
     case GI_TYPE_TAG_DOUBLE:
+    case GI_TYPE_TAG_UNICHAR:
         break;
 
     case GI_TYPE_TAG_FILENAME:
@@ -2098,6 +2552,9 @@ gjs_g_arg_release_internal(JSContext  *context,
             } else if (g_type_is_a(gtype, G_TYPE_BOXED)) {
                 if (transfer != TRANSFER_IN_NOTHING)
                     g_boxed_free(gtype, arg->v_pointer);
+            } else if (g_type_is_a(gtype, G_TYPE_VARIANT)) {
+                if (transfer != TRANSFER_IN_NOTHING)
+                    g_variant_unref (arg->v_pointer);
             } else if (gtype == G_TYPE_NONE) {
                 if (transfer != TRANSFER_IN_NOTHING) {
                     gjs_throw(context, "Don't know how to release GArgument: not an object or boxed type");
@@ -2162,7 +2619,7 @@ gjs_g_arg_release_internal(JSContext  *context,
             case GI_TYPE_TAG_FILENAME:
                 if (transfer == GI_TRANSFER_CONTAINER)
                     g_free(arg->v_pointer);
-                else if (transfer == GI_TRANSFER_EVERYTHING)
+                else
                     g_strfreev (arg->v_pointer);
                 break;
 
@@ -2175,8 +2632,35 @@ gjs_g_arg_release_internal(JSContext  *context,
                 g_free (arg->v_pointer);
                 break;
 
+            case GI_TYPE_TAG_INTERFACE:
+            case GI_TYPE_TAG_GLIST:
+            case GI_TYPE_TAG_GSLIST:
+            case GI_TYPE_TAG_ARRAY:
+            case GI_TYPE_TAG_GHASH:
+                if (transfer != GI_TRANSFER_CONTAINER) {
+                    g_assert (g_type_info_is_zero_terminated (type_info));
+                    gpointer *array;
+                    GArgument elem;
+
+                    for (array = arg->v_pointer; *array; array++) {
+                        elem.v_pointer = *array;
+                        if (!gjs_g_arg_release_internal(context,
+                                                        GI_TRANSFER_EVERYTHING,
+                                                        param_info,
+                                                        element_type,
+                                                        &elem)) {
+                            failed = JS_TRUE;
+                        }
+                    }
+                }
+                g_free (arg->v_pointer);
+                break;
+
             default:
-                g_assert_not_reached ();
+                gjs_throw(context,
+                          "Releasing a C array with explicit length, that was nested"
+                          "inside another container. This is not supported (and will leak)");
+                failed = JS_TRUE;
             }
 
             g_base_info_unref((GIBaseInfo*) param_info);
@@ -2209,7 +2693,7 @@ gjs_g_arg_release_internal(JSContext  *context,
             case GI_TYPE_TAG_GHASH:
                 if (transfer == GI_TRANSFER_CONTAINER) {
                     g_array_free((GArray*) arg->v_pointer, TRUE);
-                } else if (transfer == GI_TRANSFER_EVERYTHING) {
+                } else {
                     GArray *array = arg->v_pointer;
                     guint i;
 
@@ -2218,7 +2702,7 @@ gjs_g_arg_release_internal(JSContext  *context,
 
                         arg.v_pointer = g_array_index (array, gpointer, i);
                         gjs_g_argument_release(context,
-                                               GI_TRANSFER_EVERYTHING,
+                                               transfer,
                                                param_info,
                                                &arg);
                     }
@@ -2366,5 +2850,88 @@ gjs_g_argument_release_in_arg(JSContext  *context,
     return JS_TRUE;
 }
 
+JSBool
+gjs_g_argument_release_in_array (JSContext  *context,
+                                 GITransfer  transfer,
+                                 GITypeInfo *type_info,
+                                 guint       length,
+                                 GArgument  *arg)
+{
+    GITypeInfo *param_type;
+    gpointer *array;
+    GArgument elem;
+    guint i;
+    JSBool ret = JS_TRUE;
+    GITypeTag type_tag;
 
+    if (transfer != GI_TRANSFER_NOTHING)
+        return JS_TRUE;
 
+    gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
+                      "Releasing GArgument array in param");
+
+    array = arg->v_pointer;
+
+    param_type = g_type_info_get_param_type(type_info, 0);
+    type_tag = g_type_info_get_tag(param_type);
+
+    if (type_needs_release(param_type, type_tag)) {
+        for (i = 0; i < length; i++) {
+            elem.v_pointer = array[i];
+            if (!gjs_g_arg_release_internal(context, TRANSFER_IN_NOTHING,
+                                            param_type, type_tag, &elem)) {
+                ret = JS_FALSE;
+                break;
+            }
+        }
+    }
+
+    g_base_info_unref(param_type);
+    g_free(array);
+
+    return ret;
+}
+
+JSBool
+gjs_g_argument_release_out_array (JSContext  *context,
+                                  GITransfer  transfer,
+                                  GITypeInfo *type_info,
+                                  guint       length,
+                                  GArgument  *arg)
+{
+    GITypeInfo *param_type;
+    gpointer *array;
+    GArgument elem;
+    guint i;
+    JSBool ret = JS_TRUE;
+    GITypeTag type_tag;
+
+    if (transfer == GI_TRANSFER_NOTHING)
+        return JS_TRUE;
+
+    gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
+                      "Releasing GArgument array out param");
+
+    array = arg->v_pointer;
+
+    param_type = g_type_info_get_param_type(type_info, 0);
+    type_tag = g_type_info_get_tag(param_type);
+
+    if (transfer != GI_TRANSFER_CONTAINER) {
+        for (i = 0; i < length; i++) {
+            elem.v_pointer = array[i];
+            if (!gjs_g_arg_release_internal(context,
+                                            GI_TRANSFER_EVERYTHING,
+                                            param_type,
+                                            type_tag,
+                                            &elem)) {
+                ret = JS_FALSE;
+            }
+        }
+    }
+
+    g_base_info_unref(param_type);
+    g_free(array);
+
+    return ret;
+}
