@@ -349,19 +349,6 @@ object_instance_new_resolve(JSContext *context,
 
             method_name = g_base_info_get_name( (GIBaseInfo*) method_info);
 
-            /* we do not define deprecated methods in the prototype */
-            if (g_base_info_is_deprecated( (GIBaseInfo*) method_info)) {
-                gjs_debug(GJS_DEBUG_GOBJECT,
-                          "Ignoring definition of deprecated method %s in prototype for %s (%s.%s)",
-                          method_name,
-                          g_type_name(priv->gtype),
-                          g_base_info_get_namespace( (GIBaseInfo*) priv->info),
-                          g_base_info_get_name( (GIBaseInfo*) priv->info));
-                g_base_info_unref( (GIBaseInfo*) method_info);
-                ret = JS_TRUE;
-                goto out;
-            }
-
             gjs_debug(GJS_DEBUG_GOBJECT,
                       "Defining method %s in prototype for %s (%s.%s)",
                       method_name,
